#!/usr/bin/env python

import os
from pathlib import Path
from azure.ai.ml import MLClient
from azure.identity import ManagedIdentityCredential
import logging
from typing import Tuple
import yaml
from opencensus.ext.azure.log_exporter import AzureLogHandler

def client(
    logging_level: int=logging.ERROR,
    logging_disabled: bool=False,
    workspace_yaml: str=os.path.join(Path(__file__).parent, "workspace.yaml"),
    manual_yaml_file_path: str="",
    mode: str="file",
    app_insight_connection: str=None
) -> Tuple[logging.getLogger, MLClient]:
    """Initiate the logger and AML client for a given workspace

    Create the logger at a desired level, connected to a pre-configured application insights and log
    analytics workspace. Connected to these resources via Instrumentation Key. Also return a workspace-enable AML
    client object.

    Args: 
        logging_level (int) : severity level for python logging according to docs at https://docs.python.org/3/library/logging.html#logging-levels
        logging_disabled (bool) : disable logging argument, default level is ERROR
        workspace_yaml (str) : path to workspace yaml location for authentication
        manual_yaml_file_path (str) : path to workspace yaml location
        mode (str) : whether or not to execute command as script or in databricks
        app_insight_connection (str) : connection string for application insights
    """

    
    # configure logging level
    logging.basicConfig(level=logging_level)
    # set logging level
    logger = logging.getLogger(__name__)
    # TODO: replace the all-zero GUID with your instrumentation key.
    if app_insight_connection is not None:
        logger.addHandler(AzureLogHandler(
            connection_string=app_insight_connection)
        )
    

    logger.info("Logger handler added.")
    # retrieve default aml network credential
    credential = ManagedIdentityCredential(client_id=os.environ.get("DEFAULT_IDENTITY_CLIENT_ID", None))

    try: 
        ml_client = MLClient.from_config(credential)
    except Exception as exc:
        # print exception for attempted authorization
        logger.error(exc)
        # get client handle to the workspace if default authentication fails
        # load in workspace config
        if mode=='file':
            file_path = os.path.join(Path(__file__).parent, workspace_yaml)
        else:
            assert manual_yaml_file_path!="", ValueError("set manual_file_path argument")
            file_path = os.path.join(manual_yaml_file_path, workspace_yaml)
            
        with open(file_path, "r") as stream:
            try:
                ws_config = yaml.safe_load(stream)
            except yaml.YAMLError as exc:
                logger.error(exc)
        # add default credential to workspace config
        ws_config['credential'] = credential
        ml_client = MLClient(**ws_config)

    # get workspace details from client
    workspace = ml_client.workspaces.get(name=ml_client.workspace_name)
    subscription_id = ml_client.connections._subscription_id
    resource_group = workspace.resource_group
    workspace_name = ml_client.workspace_name
    location = workspace.location
    
    # disable logger if requested
    if logging_disabled:
        logger = logging.getLogger(__name__)
        logger.propagate = False
    # push workspace details to logs
    logger.info(f"{subscription_id}:{resource_group}:{workspace_name}:{location}")
    return logger, ml_client, credential




    


# imports
import os
import mlflow

from random import random

# define functions
def main():
    mlflow.log_param("example_param", "example")
    mlflow.log_metric("example_metric", random())
    os.system(f"echo 'hello world' > helloworld.txt")
    mlflow.log_artifact("helloworld.txt")


# run functions
if __name__ == "__main__":
    # run main function
    main()


#!/usr/bin/env python

# import the libraries
from typing import Optional, Dict
from azure.ai.ml import command
from azure.ai.ml import Input
from ._initialization import client

class AzureMLJob:
    """Job submission class for azureml command to run against data assets

    A temporary wrapper class for sending script runs, complete with mlflow logging, 
    to a specified experiment and job bucket.

    Params:
        input_data (str) : path to data in azureml datastores
        input_configuration (Dict) : config struct optionally passed by Hydra
        tags (Dict) : tag dictionary
        code_directory (str) : directory location on compute instance of codebase, e.g. {compute_instance}/code/Models/test/src/
        command (str) : command to run in directory, complete with dict specifications
        environment (str) : environment logged to azure upon which to run codebase command
        compute_cluster (str) : compute resources in azure to run jobs
        job (str) : display name of job
        experiment (str) : display name of experiment
    """

    def __init__(
        self, 
        input_data: Optional[str]= None,
        input_configuration: Optional[Dict] = None, 
        tags: Optional[Dict] = None, 
        code_directory: Optional[str] = None,
        cli_command: Optional[str] = None, 
        environment: Optional[str] = None,
        compute_cluster: Optional[str] = None,
        job: Optional[str] = None, 
        experiment: Optional[str] = None,
        **kwargs
    ):
        logger, ml_client, credential = client()
        # set ml_client and experiment name
        self._ml_client = ml_client
        self.experiment = experiment
        # point input data into configuration dict
        input_configuration['data'] = Input(
            type="uri_file", 
            path=input_data
        )
        # clean code directory string 
        code_directory_strip = code_directory.lstrip("/")
        codebase = "/mnt/batch/tasks/shared/LS_root/mounts/clusters/"+code_directory_strip
        # construct command from specification
        self.command = command(
                inputs=input_configuration,
                code=codebase,  # location of source code
                # The inputs/outputs are accessible in the command via the ${{ ... }} notation
                command=cli_command,
                # ready-made environment
                environment=environment,
                compute=compute_cluster,
                # An experiment is a container for all the iterations one does on a certain project. All the jobs submitted under the same experiment name would be listed next to each other in Azure ML studio.
                experiment_name=experiment,
                display_name=job, 
                tags=tags
            )
        
    @property 
    def command(self):
        return self._command
    
    @command.setter
    def command(self, value):
        self._command = value

    def run(self):
        """Create or update job run under experiment name

        Params: 
            self (AzureMLJob) : object
        """
        command_to_run = self.command
        # set job on create or update
        client_job = self._ml_client.jobs.create_or_update(command_to_run)
        return client_job


# # configure the command job
# job = command(
#     inputs=dict(
#         # uri_file refers to a specific file as a data asset
#         data=Input(
#             type="uri_file",
#             path="https://azuremlexamples.blob.core.windows.net/datasets/credit_card/default%20of%20credit%20card%20clients.csv",
#         ),
#         test_train_ratio=0.2,  # input variable in main.py
#         learning_rate=0.25,  # input variable in main.py
#         registered_model_name=registered_model_name,  # input variable in main.py
#     ),
#     code="/mnt/batch/tasks/shared/LS_root/mounts/clusters/cpaulis2-ci/code/Models/test/src/",  # location of source code
#     # The inputs/outputs are accessible in the command via the ${{ ... }} notation
#     command="python main.py --data ${{inputs.data}} --test_train_ratio ${{inputs.test_train_ratio}} --learning_rate ${{inputs.learning_rate}} --registered_model_name ${{inputs.registered_model_name}}",
#     # This is the ready-made environment you are using
#     environment="AzureML-sklearn-1.0-ubuntu20.04-py38-cpu@latest",
#     # This is the compute you created earlier
#     compute="dualnode1-cc",
#     # An experiment is a container for all the iterations one does on a certain project. All the jobs submitted under the same experiment name would be listed next to each other in Azure ML studio.
#     experiment_name="train_model_credit_default_prediction",
#     display_name="credit_default_prediction",
# )




#!/usr/bin/env python

import os
from time import sleep
from typing import Union
import io
from time import sleep
import pandas as pd
from azure.ai.ml.entities import Data
from typing import Optional, Dict, Any
from azure.ai.ml.constants._common import AssetTypes
from azure.identity import ManagedIdentityCredential, DefaultAzureCredential
from azure.keyvault.secrets import SecretClient
from azure.storage.blob import BlobClient, DelimitedTextDialect, ContainerClient
from ._initialization import client


class AzureMLDataAsset(Data):
    """Data wrapper class for azureml data assets
    
    A child class of the SDKv2 AzureML data class to expedite the creation and maintenance of 
    data assets for workflow assignments and pipeline jobs.

    Params:
        name (str) : Name of the resource
        version (str) : version of the resource
        description (str) : description of the resource
        tags (Dict) : dictionary tags for the resource
        properties (Dict) : asset properties dictionary from artifact class https://github.com/Azure/azure-sdk-for-python/blob/azure-ai-ml_1.3.0/sdk/ml/azure-ai-ml/azure/ai/ml/entities/_assets/asset.py
        path (str) : path to access resource

    """
    def __init__(
        self,
        *,
        description: Optional[str] = None,
        tags: Optional[Dict] = None,
        properties: Optional[Dict] = None,
        path: Optional[str] = None,
        type: str = AssetTypes.URI_FOLDER,
        **kwargs,
    ):
        # replicate init function from data class from AzureML sdk documentation
        logger, ml_client, credential = client()
        self._ml_client = ml_client
        self._skip_validation = kwargs.pop("skip_validation", False)
        self._mltable_schema_url = kwargs.pop("mltable_schema_url", None)
        self._referenced_uris = kwargs.pop("referenced_uris", None)
        self.type = type
        
        # set name equal to file name (sans suffix) from path
        self.name = path.split("/")[-1].split(".")[0]
        super().__init__(
            name=self.name,
            path=path,
            description=description,
            tags=tags,
            type=type,
            properties=properties,
            **kwargs,
        )
        self.path = path
        try:
            latest_version = {d.name: d.latest_version for d in ml_client.data.list()}[self.name]
            self.latest_version = latest_version
        except KeyError as e:
            logger.error(e)
            logger.info("Data asset does not exist. Creating first version.")
            self.latest_version = "1"

    def _get_all_latest_assets(self) -> Dict[str, str]:
        """Get all data assets registered in workspace

        Params:
            self (AzureMLDataAsset) : object
        """
        # use client to retrieve available assets and asset latest versions
        asset_versions = {d.name: d.latest_version for d in self._ml_client.data.list()}
        return asset_versions
    
    def _get_latest_version(self) -> int:
        """Get latest version for specified asset

        Params:
            self (AzureMLDataAsset) : object
        """
        # use client to retrieve available assets and asset latest versions
        asset_version = self._get_all_latest_assets()[self.name]
        return asset_version

    @property 
    def latest_version(self):
        return self._latest_version
    
    @latest_version.setter
    def latest_version(self, value):
        self._latest_version = value

    def update_to_new_version(self) -> Any:
        """Create resource if this currently does not exist.
        
        Params:
            self (AzureMLDataAsset) : object 
        """
        # check to see if current object exists
        if self.name not in self._get_all_latest_assets().keys():
            # create data resource if not extant
            self.latest_version = "1"
            self._ml_client.data.create_or_update(self)
            return self
        else:
            self._ml_client.data.create_or_update(self)
            sleep(3)
            self.latest_version = self._get_latest_version()
            return self

def df_from_delimited_blob(
    blob_url: str, 
    credential: Union[DefaultAzureCredential, ManagedIdentityCredential], 
    **kwargs: Any    
) -> pd.DataFrame:
    """Returns a pandas dataframe from a blob URL

    Args: 
        blob_url (str): blob url string including account and container names on url path
        credential (Union[DefaultAzureCredential, ManagedIdentityCredential]): auth credential for Azure Blob Client

    Returns: 
        df (pd.DataFrame): returned dataframe decoded from bytes

    """
    # retrieve keyvault from os environment vars
    assert "KEY_VAULT_NAME" in os.environ, KeyError("KEY_VAULT_NAME must be set in environment vars to access storage token.")
    keyVaultName = os.environ["KEY_VAULT_NAME"]
    KVUri = f"https://{keyVaultName}.vault.azure.net"
    # get kv client from credential
    kv_client = SecretClient(vault_url=KVUri, credential=credential)
    storage_account = blob_url.split("/")[2].split(".")[0]
    retrieved_secret = kv_client.get_secret(f"{storage_account}-access-key-1")
    # parse blob url for storage account url, container, and blob 
    blob_url_split = blob_url.split("/")
    account_url = "/".join(blob_url_split[:3])+"/"
    container = blob_url_split[3]
    blob = "/".join(blob_url_split[4:])
    # connect to blob through client
    blob_client = BlobClient(account_url=account_url, credential=retrieved_secret.value, container_name=container, blob_name=blob)
    # specify IO format (include header)
    input_format = DelimitedTextDialect(delimiter=',', quotechar='"', lineterminator='\n', escapechar="", has_header=True)
    output_format = DelimitedTextDialect(delimiter=',', quotechar='"', lineterminator='\n', escapechar="", has_header=True)
    reader = blob_client.query_blob("SELECT * from BlobStorage", blob_format=input_format, output_format=output_format)
    # read into bytes-like object
    content = reader.readall()
    # decode dataframe from bytes
    df = pd.read_csv(io.StringIO(content.decode('utf-8')))
    return df


def df_to_delimited_blob(
    df: pd.DataFrame,
    storage_account_url: str,
    output_blob: str, 
    credential: Union[DefaultAzureCredential, ManagedIdentityCredential], 
    **kwargs: Any    
) :
    """uploads a pandas data frame to blob storage
    
    Args: 
        df (pd.DataFrame) : pandas dataframe to upload to blob location
        blob_url (str): blob url string including account and container names on url path
        credential (Union[DefaultAzureCredential, ManagedIdentityCredential]): auth credential for Azure Blob Client

    Returns: 
        blob_props (pd.DataFrame): returned properties of blob when created
    """
    assert "KEY_VAULT_NAME" in os.environ, KeyError("KEY_VAULT_NAME must be set in environment vars to access storage token.")
    keyVaultName = os.environ["KEY_VAULT_NAME"]
    KVUri = f"https://{keyVaultName}.vault.azure.net"
    # get kv client from credential
    kv_client = SecretClient(vault_url=KVUri, credential=credential)
    storage_account = storage_account_url.split("/")[2].split(".")[0]
    retrieved_secret = kv_client.get_secret(f"{storage_account}-access-key-1")
    # link to blob client
    output_blob_split = output_blob.split("/")
    container = output_blob_split[0]
    blob = "/".join(output_blob_split[1:])
    container_client = ContainerClient(account_url=storage_account_url, credential=retrieved_secret.value, container_name=container)
    # convert pandas dataframe to string
    output = io.StringIO()
    output = df.to_csv(encoding='utf-8', index=False)
    # try uploading blob
    try:
        container_client.upload_blob(blob, output, overwrite=True, encoding='utf-8')
    except Exception as err:
        print (f"Unexpected blob upload error {err=}, {type(err)=}")
    # get blob props
    sleep(3)
    blob_client = BlobClient(account_url=storage_account_url, credential=retrieved_secret.value, container_name=container, blob_name=blob)
    blob_props = blob_client.get_blob_properties()
    return blob_props
    

    



from setuptools import setup
setup(
    name='idpaml',
    version='1.0.2',
    author='cpaulis2',
    description='AzureML wrapper for disease modeling.',
    long_description='This is a longer description for the project',
    url='https://github.com/uhg-internal/idp-aml',
    keywords='azureml, idp, aml',
    python_requires='>=3.6, <4',
    install_requires=[
        "azure-ai-ml",
        "azure-identity",
        "azure-storage-blob<=12.13.0",
        "hydra-core",
        "opencensus-ext-azure",
        "PyYAML",
        "pyfiglet",
        "pandas",
        "pickle5",  # imported as pickle
        "pyyaml",   # imported as yaml
        ]
)


# Handle to the workspace
from azure.ai.ml import MLClient
# Authentication package
from azure.identity import DefaultAzureCredential
import logging
from opencensus.ext.azure.log_exporter import AzureLogHandler

logger = logging.getLogger(__name__)
credential = DefaultAzureCredential()

# Get a handle to the workspace
ml_client = MLClient(
    credential=credential,
    subscription_id="b7fbc2be-be14-4cea-9633-40351430064d",
    resource_group_name="aml-starter-rg-dev",
    workspace_name="dispred-aml",
)
cpu_compute_target = "dualnode1-cc"
# let's see if the compute target already exists
cpu_cluster = ml_client.compute.get(cpu_compute_target)
logger.warning("THIS IS A TEST WARNING LOG.")
print(
    f"You already have a cluster named {cpu_compute_target}, we'll reuse it as is."
)


from azureml.core import Workspace, Environment, Experiment, ScriptRunConfig
from azureml.core.conda_dependencies import CondaDependencies

import mlflow

ws = Workspace.from_config()
mlflow.set_tracking_uri(ws.get_mlflow_tracking_uri())
experiment_name = 'example-experiment'
mlflow.set_experiment(experiment_name)

# env = Environment(name="mlflow-env")
# # Specify conda dependencies with scikit-learn and temporary pointers to mlflow extensions
# cd = CondaDependencies.create(
#     conda_packages=["scikit-learn", "matplotlib"],
#     pip_packages=["azureml-mlflow", "pandas", "numpy"]
#     )


# configure script to run on compute
src = ScriptRunConfig(source_directory="src",
                      script="train.py",
                      compute_target="singlenode1-cc",
                      environment='azureml_py310_sdkv2')

exp = Experiment(workspace=ws, name=experiment_name)
run = exp.submit(src)